#!/usr/bin/env python

# The software in this package is distributed under the GNU General
# Public License version 2 (with a special exception described below).
#
# A copy of GNU General Public License (GPL) is included in this distribution,
# in the file COPYING.GPL.
#
# As a special exception, if other files instantiate templates or use macros
# or inline functions from this file, or you compile this file and link it
# with other works to produce a work based on this file, this file
# does not by itself cause the resulting work to be covered
# by the GNU General Public License.
#
# However the source code for this file must still be made available
# in accordance with section (3) of the GNU General Public License.
#
# This exception does not invalidate any other reasons why a work based
# on this file might be covered by the GNU General Public License.
#
# Copyright (c) 2016-2018 Intra2net AG <info@intra2net.com>

"""Unit test for build_cnfvar.py"""
import unittest
import os
from src.cnfline.build_cnfvar import BuildCnfVar

TEST_CONFIG_FILENAME = 'some_config.cnf'


class BuildCnfVarTest(unittest.TestCase):
    def tearDown(self):
        if os.path.isfile(TEST_CONFIG_FILENAME):
            os.unlink(TEST_CONFIG_FILENAME)

    def test_simple(self):
        cnfvar = BuildCnfVar('FOOBAR', 123, 'some_data', 10)

        self.assertEqual('FOOBAR', cnfvar.name)
        self.assertEqual(123, cnfvar.instance)
        self.assertEqual('some_data', cnfvar.data)
        self.assertEqual(10, cnfvar.line_no)

    def test_find_child_line_no(self):
        cnfvar = BuildCnfVar('FOOBAR', 123, 'some_data', 10)
        new_child_line_no = cnfvar.add_cnf('FOOBAR_CHILD', 0, 'xxx')

        self.assertEqual(new_child_line_no,
                         cnfvar.find_child_line_no('FOOBAR_CHILD'))

        # Should not be found
        self.assertEqual(0, cnfvar.find_child_line_no('FOOBAR_SPACESHARK'))

    def test_find_free_line_no(self):
        cnfvar = BuildCnfVar('FOOBAR', 123, 'some_data', 10)
        self.assertEqual(11, cnfvar.find_free_line_no())

        cnfvar.add_cnf('FOOBAR_CHILD', 0, 'xxx')
        self.assertEqual(12, cnfvar.find_free_line_no())

    def test_find_free_child_instance(self):
        cnfvar = BuildCnfVar('FOOBAR', 123, 'some_data', 10)
        cnfvar.add_cnf('FOOBAR_CHILD', 0, 'xxx')

        self.assertEqual(0, cnfvar.find_free_child_instance('FOOBAR_OTHER'))
        self.assertEqual(1, cnfvar.find_free_child_instance('FOOBAR_CHILD'))

    def test_update_cnf(self):
        cnfvar = BuildCnfVar('FOOBAR', 123, 'some_data', 10)
        cnfvar.add_cnf('FOOBAR_CHILD', 0, 'xxx')

        # Update existing cnfvar
        cnfvar.update_cnf('FOOBAR_CHILD', 0, 'abc')

        self.assertEqual('10 FOOBAR,123: "some_data"\n'
                         '11 (10) FOOBAR_CHILD,0: "abc"\n', str(cnfvar))

    def test_string_output(self):
        cnfvar = BuildCnfVar('FOOBAR', 123, 'some_data', 10)
        cnfvar.add_cnf('FOOBAR_CHILD', 0, 'xxx')
        cnfvar.update_cnf('FOOBAR_CHILD', 0, 'abc')
        cnfvar.add_cnf('FOOBAR_CHILD', 1, 'more data')

        self.assertEqual('10 FOOBAR,123: "some_data"\n'
                         '11 (10) FOOBAR_CHILD,0: "abc"\n'
                         '12 (10) FOOBAR_CHILD,1: "more data"\n', str(cnfvar))

    def test_del_cnf(self):
        cnfvar = BuildCnfVar('FOOBAR', 123, 'some_data', 10)
        cnfvar.add_cnf('FOOBAR_CHILD', 0, 'xxx')
        cnfvar.add_cnf('FOOBAR_CHILD', 1, 'more data')

        cnfvar.del_cnf('FOOBAR_CHILD')

        self.assertEqual('10 FOOBAR,123: "some_data"\n', str(cnfvar))

    def test_add_different_parent_no(self):
        cnfvar = BuildCnfVar('FOOBAR', 123, 'some_data', 10)
        sub_parent = cnfvar.add_cnf('FOOBAR_CHILD', 0, 'xxx')
        cnfvar.add_cnf('FOOBAR_OTHER', 0, 'foo')
        cnfvar.add_cnf('FOOBAR_CHILD_TYPE', 1, 'spaceshark', sub_parent)
        cnfvar.add_cnf('FOOBAR_OTHER2', 0, 'foo2')

        self.assertEqual('10 FOOBAR,123: "some_data"\n'
                         '11 (10) FOOBAR_CHILD,0: "xxx"\n'
                         '12 (10) FOOBAR_OTHER,0: "foo"\n'
                         '13 (11) FOOBAR_CHILD_TYPE,1: "spaceshark"\n'
                         '14 (10) FOOBAR_OTHER2,0: "foo2"\n', str(cnfvar))

    def test_add_defaults(self):
        cnfvar = BuildCnfVar('FOOBAR', 0, 'some_data')

        defaults = {'FOOBAR_SOMETHING': 'abc',
                    'FOOBAR_MODE': 'optimize'}
        cnfvar.add_defaults(defaults)

        self.assertTrue('1 FOOBAR,0: "some_data"\n'
                        '2 (1) FOOBAR_SOMETHING,0: "abc"\n'
                        '3 (1) FOOBAR_MODE,0: "optimize"\n' == str(cnfvar)
                     or '1 FOOBAR,0: "some_data"\n'
                        '2 (1) FOOBAR_MODE,0: "optimize"\n'
                        '3 (1) FOOBAR_SOMETHING,0: "abc"\n' == str(cnfvar))

    def test_mark_as_own_parent(self):
        cnfvar = BuildCnfVar('FOOBAR_SOMETHING', 123, 'some_data', 10)

        line_no = cnfvar.add_cnf('FOOBAR_OTHER', 0, 'xxx')
        cnfvar.mark_as_own_parent(line_no)

        self.assertEqual('10 FOOBAR_SOMETHING,123: "some_data"\n'
                         '11 FOOBAR_OTHER,0: "xxx"\n', str(cnfvar))

    def test_save(self):
        cnfvar = BuildCnfVar('FOOBAR', 0, 'some_data')

        defaults = {'FOOBAR_SOMETHING': 'abc',
                    'FOOBAR_MODE': 'optimize'}
        cnfvar.add_defaults(defaults)

        cnfvar.save(TEST_CONFIG_FILENAME)
        with open(TEST_CONFIG_FILENAME, 'r') as input:
            read_back = input.read()

            self.assertTrue('1 FOOBAR,0: "some_data"\n'
                            '2 (1) FOOBAR_SOMETHING,0: "abc"\n'
                            '3 (1) FOOBAR_MODE,0: "optimize"\n' == read_back
                         or '1 FOOBAR,0: "some_data"\n'
                            '2 (1) FOOBAR_MODE,0: "optimize"\n'
                            '3 (1) FOOBAR_SOMETHING,0: "abc"\n' == read_back)

        os.unlink(TEST_CONFIG_FILENAME)


if __name__ == '__main__':
    unittest.main()
